(declaim (optimize (speed 3) (safety 1)))

(define-object *primitive-nil*
  (traits nil t))

(define-object *traits-method*
  (name "Method" nil)
  (traits nil t)
  (cloneable *primitive-nil* t))
(define-object *primitive-method*
  (traits *traits-method* t)
  ("lineNumber" 0 nil)
  ("fileName" "" nil)
  (selector nil nil)
  (accessor nil)
  (arguments '())
  ("sourceCode" nil nil)
  (code nil))
;Implemented using *primitive-nil* and *primitive-method*
(add-accessors *traits-method* '|cloneable| t)
(add-accessors *traits-method* '|traits| t)
(add-accessors *traits-method* '|name| t)
(add-accessors *primitive-method* '|traits| t)
(add-accessors *primitive-method* '|sourceCode| t)
(add-accessors *primitive-method* '|lineNumber| t)
(add-accessors *primitive-method* '|fileName| t)
(add-accessors *primitive-nil* '|traits| t)

(defmacro define-named-object (name string-name &rest rest)
  `(define-object ,name (name ,string-name nil :immutable) ,@rest))

(define-method "argumentNames" ((method *traits-method*))
  (coerce (get-slot method '|arguments|) 'vector))

(define-named-object *traits-root* "Root"
  (traits nil t :immutable))
(define-object *primitive-root*
  (traits *traits-root* t :immutable))

(define-named-object *traits-traits* "Traits"
  (root *traits-root* t :immutable))
(define-named-object *primitive-traits* "_"
  (traits *traits-traits* t :immutable))
(set-slot *traits-method* '|traits| *traits-traits*)
(set-slot *traits-root* '|traits| *traits-traits*)

(defmacro define-traits (name string-name &rest slots)
  `(progn
     (define-named-object ,name ,string-name
       (traits *traits-traits* t :immutable) ,@slots)
     (define-method traits ((obj ,name))
       (if (eq obj ,name)
           *traits-traits*
           ,name))))

(define-traits *traits-oddball* "Oddball"
  (root *traits-root* t :immutable))
(define-object *primitive-oddball*
  (traits *traits-oddball* t :immutable))
(set-slot *primitive-nil* '|traits| *traits-oddball*)

(defmacro define-named-oddball (name string-name &rest slots)
  `(define-named-object ,name ,string-name
     (traits *traits-oddball* t :immutable) ,@slots))

(define-named-oddball *no-role* "NoRole")

(define-traits *traits-derivable* "Derivable"
  (root *traits-root* t :immutable))
(define-object *primitive-derivable*
  (traits *traits-derivable* t :immutable))

(define-traits *traits-cloneable* "Cloneable"
  (derivable *traits-derivable* t :immutable))
(define-object *primitive-cloneable*
  (traits *traits-cloneable* t :immutable))
(set-slot *traits-method* '|cloneable| *traits-cloneable*)

(defmacro bool-of (&body body)
  `(if (progn ,@body) *primitive-true* *primitive-false*))

(defmacro define-resource-method (name arguments &body body)
  (let ((socket-sym (first (first arguments))))
    `(define-method ,name ,arguments
       (let ((,socket-sym (if (wm-object-p ,socket-sym)
			      (get-slot ,socket-sym '|handle|)
			    ,socket-sym)))
	 ,@body))))

(define-traits *traits-lisp* "Lisp"
  (oddball *traits-oddball* t :immutable))

(define-traits *traits-boolean* "Boolean"
  (oddball *traits-oddball* t :immutable))
(define-object *primitive-true*
  (traits *traits-boolean* t :immutable))
(define-object *primitive-false*
  (traits *traits-boolean* t :immutable))

(define-traits *traits-number* "Number"
  (derivable *traits-derivable* t :immutable))
(define-object *primitive-number*
  (traits *traits-number* t :immutable))

(define-traits *traits-integer* "Integer"
  (number *traits-number* t :immutable))

(define-traits *traits-float* "Float"
  (number *traits-number* t :immutable))

(define-traits *traits-character* "Character"
  (oddball *traits-oddball* t :immutable))

(define-traits *traits-symbol* "Symbol"
  (oddball *traits-oddball* t :immutable))

(define-traits *traits-array* "Array"
  (collection *traits-cloneable* t :immutable))

(define-traits *traits-byte-array* "ByteArray"
  (collection *traits-cloneable* t :immutable))

(define-traits *traits-word-array* "WordArray"
  (collection *traits-cloneable* t :immutable))

(define-traits *traits-string* "String"
  (collection *traits-cloneable* t :immutable))

(define-traits *traits-file* "FileHandle"
  (parent *traits-oddball* t t))
(define-object *primitive-file*
  (traits *traits-file* t :immutable))

#+clisp
(progn					; CLISP sockets.
(define-traits *traits-socket-client* "SocketClientHandle"
  (oddball *traits-oddball* t :immutable))
(define-object *primitive-socket-client*
  (traits *traits-socket-client* t :immutable))
(define-traits *traits-socket-server* "SocketServerHandle"
  (oddball *traits-oddball* t :immutable))
(define-object *primitive-socket-server*
  (traits *traits-socket-server* t :immutable))
)

(define-traits *traits-namespace* "Namespace"
  (oddball *traits-oddball* t :immutable))
(define-object *primitive-namespace*
  (traits *traits-namespace* t :immutable))

#+subjective-slate
(progn
  (define-traits *traits-subject* "Subject"
    (cloneable *traits-cloneable* t t))

  (define-object *primitive-subject*
    (traits *traits-subject* t :immutable)
    ("parentSubject" *primitive-nil* t t)
    (focus *primitive-nil* t t)))

#+layered-slate
(progn
  (define-traits *traits-layer* "Layer"
    (cloneable *traits-cloneable* t))

  (define-object *primitive-layer*
    (traits *traits-layer* t :immutable)
    ("parentLayer" *primitive-nil* t t)
    (focus *primitive-nil* t t)))

(define-named-oddball *primitive-literal* "PrimitiveLiteral")

(define-traits *traits-backtrace-frame* "BacktraceFrame"
  (cloneable *traits-cloneable* t))

(define-object *primitive-backtrace-frame*
  (traits *traits-backtrace-frame* t :immutable)
  (method *primitive-nil* nil t)
  (selector *primitive-nil* nil t)
  (arguments *primitive-nil* nil t)
  (optionals *primitive-nil* nil t))

(define-object *prototypes*
  (traits *traits-namespace* t :immutable)
  ("Traits" *primitive-traits* nil :immutable)
  ("Root" *primitive-root* nil :immutable)
  ("Cloneable" *primitive-cloneable* nil :immutable)
  ("Oddball" *primitive-oddball* nil :immutable)
  ("Derivable" *primitive-derivable* nil :immutable)
  ("Namespace" *primitive-namespace* nil :immutable)
  ("Nil" *primitive-nil* nil :immutable)
  ("Boolean" *primitive-true* nil :immutable)
  ("True" *primitive-true* nil :immutable)
  ("False" *primitive-false* nil :immutable)
  ("Number" *primitive-number* nil :immutable)
  ("Integer" 0 nil :immutable)
  ("Float" 0.0 nil :immutable)
  ("ASCIICharacter" #+cormanlisp #\Nul #-cormanlisp #\Null nil :immutable)
  ("Symbol" '|symbol| nil :immutable)
  ("Array" (make-array 0 :initial-element *primitive-nil*) nil :immutable)
  ("ByteArray" (make-array 0 :element-type '(unsigned-byte 8)) nil :immutable)
  ("WordArray" (make-array 0 :element-type '(unsigned-byte 32)) nil :immutable)
  ("String" "" nil :immutable)
  ("FileHandle" *primitive-file* nil :immutable)
  #+clisp ("SocketClientHandle" *primitive-socket-client* nil :immutable)
  #+clisp ("SocketServerHandle" *primitive-socket-server* nil :immutable)
  ("Lisp" (lambda () *primitive-nil*) nil :immutable)
  ("BacktraceFrame" *primitive-backtrace-frame* nil :immutable)
  #+subjective-slate ("Subject" *primitive-subject* nil :immutable)
  #+layered-slate ("Layer" *primitive-layer* nil :immutable)
  ("Method" *primitive-method* nil :immutable))

(define-named-oddball *console-input* "ConsoleInput")
(define-named-oddball *console-output* "ConsoleOutput")
(define-named-oddball *console* "Console")
(define-named-oddball *platform* "Platform")

; Later replaced by an ExtensibleSequence.
(defvar *slate-modules*
  (make-array 12 :initial-element nil :fill-pointer 0 :adjustable t))

; Later replaced by an ExtensibleSequence.
(defvar *slate-features*
  (make-array 12 :initial-element nil :fill-pointer 0 :adjustable t))

#+clisp
(progn
(define-named-oddball *shell* "Shell")
(define-named-oddball *environment* "Environment")
)

(define-object *globals*
  (traits *traits-namespace* t :immutable)
  ("NoRole" *no-role* nil :immutable)
  #+clisp ("Shell" *shell* nil :immutable)
  #+clisp ("Environment" *environment* nil :immutable)
  ("Console" *console* nil :immutable)
  ("ConsoleInput" *console-input* nil :immutable)
  ("ConsoleOutput" *console-output* nil :immutable)
  ("modules" *slate-modules* nil t)
  ("features" *slate-features* nil t)
  ("currentModule" *primitive-nil* nil t)
  ("platform" *platform* nil :immutable))

(define-object *lobby*
  (traits *traits-namespace* t :immutable)
  (lobby *primitive-nil* nil :immutable)
  (prototypes *prototypes* t :immutable)
  (globals *globals* t :immutable))

(set-slot *lobby* '|lobby| *lobby*)

#+subjective-slate
(progn
  (defvar *subject* nil)

  (define-method "withoutSubject" ((method *traits-method*))
    (let ((old-subject *subject*))
      (prog2
        (setf *subject* nil)
        (apply-method method)
        (setf *subject* old-subject))))
        
  (define-method "seenFrom:" ((method *traits-method*) focus)
    (let ((new-subject (clone-object *primitive-subject*))
          (old-subject *subject*))
      (set-slot new-subject '|focus| focus)
      (set-slot new-subject '|parentSubject| old-subject)
      (prog2
        (setf *subject* new-subject)
        (apply-method method)
        (setf *subject* old-subject)))))

#+layered-slate
(progn
  (defvar *layer* nil)

  (define-method "withoutLayers" ((method *traits-method*))
    (let ((old-layer *layer*))
      (prog2
        (setf *layer* nil)
        (apply-method method)
        (setf *layer* old-layer))))

  (define-method "layering:" ((method *traits-method*) focus)
    (let ((new-layer (clone-object *primitive-layer*))
          (old-layer *layer*))
      (set-slot new-layer '|focus| focus)
      (set-slot new-layer '|parentLayer| old-layer)
      (prog2
        (setf *layer* new-layer)
        (apply-method method)
        (setf *layer* old-layer)))))

(unless (fboundp 'exit)
  (setf (fdefinition 'exit) #'quit))
(define-method "exit" ((lobby *lobby*))
  (exit))
(define-method "quit" ((lobby *lobby*))
  (exit))

(define-method "saveImageNamed:" ((lobby *lobby*) filename)
  (bool-of
      #+clisp
      (ext:saveinitmem filename :quiet t :init-function #'repl)
      #+ecl
      (progn
	(setf compiler::*cc-flags* "-g -I.")
	(c:build-program
	 filename :lisp-files +files+ :ld-flags '()
	 :epilogue-code "funcall(1,_intern(\"REPL\",user_package));"))
      #+openmcl
      (progn
	(setf ccl::*restore-lisp-functions* '(repl))
	(ccl::save-application (format nil "~A.image" filename)))
      #+cormanlisp
      (save-application filename :console t :start-function #'repl :static t)
      #+cmu
      (save-lisp (format nil "~A.core" filename) :init-function #'repl
		 :print-herald nil :site-init nil :load-init-file nil)
      #+sbcl
      (save-lisp-and-die (format nil "~A.core" filename) :toplevel #'repl)
      ))

#+clisp
(progn					; Shell-related methods.
(define-method "runProgram:withArgs:" ((shell *shell*) prog-name args)
  (apply 'ext:execute `(,prog-name ,@(coerce args 'list))))
(define-method "execute:" ((shell *shell*) command-string)
  (ext:shell command-string))
(define-method "enter" ((shell *shell*))
  (ext:shell))
(define-method "run:withArgs:inputFrom:outputTo:"
  ((shell *shell*) command-string args in out)
  (apply 'ext:run-shell-command
	 `(,command-string ,@(coerce args 'list) :input ,in :output ,out)))
(define-method "at:" ((environment *environment*) env-string)
  (ext:getenv env-string))
(define-method "at:put:" ((environment *environment*) env-string value-string)
  (setf (ext:getenv env-string) value-string))
(define-method "keys" ((environment *environment*))
  (coerce
   (loop for pair in (ext:getenv)
     collect (car pair))
   'vector))
(define-method "values" ((environment *environment*))
  (coerce
   (loop for pair in (ext:getenv)
     collect (cdr pair))
   'vector))
(define-traits *traits-pipe* "Pipe"
  (traits *traits-oddball* t t))
(define-object *pipe-primitive*
  (traits *traits-pipe* t :immutable)
  (command "" nil))
(define-method "newForCommand:" ((pipe *traits-pipe*) command-string)
  ())
)

; Only here for bootstrapping the libraries.
(define-method "requires:" ((lobby *lobby*) (features *traits-array*))
  (loop for required across features
    when (notany (lambda (present) (eq required present)) *slate-features*)
    do (format t "Feature ~A missing for module ~A.~%"
	       required (get-slot *globals* '|currentModule|)))
  *primitive-nil*)

; Only here for bootstrapping the libraries.
(define-method "provides:" ((lobby *lobby*) (features *traits-array*))
  (loop for provided across features
    unless (some (lambda (present) (eq provided present)) *slate-features*)
    do (vector-push-extend provided *slate-features*))
  *slate-features*)

; Only here for bootstrapping the libraries.
(define-method "enterModule:" ((lobby *lobby*) name)
  (vector-push-extend name *slate-modules*)
  (setf (get-slot *globals* '|currentModule|) name))

(define-method "debug" ((lobby *lobby*))
  (bool-of *slate-debug*))

(define-method "debug:" ((lobby *lobby*) (flag *traits-boolean*))
  (bool-of (setf *slate-debug* flag)))

(define-method "bits" ((float *traits-float*))
  (multiple-value-bind (significand exponent sign) (integer-decode-float float)
    (loop until (or (logbitp 0 significand) (= exponent 127))
      do (setf significand (ash significand -1))
         (incf exponent))
    (logior (if (< sign 0) (ash 1 31) 0)
            (ash significand 23)
            (+ exponent 127))))

(define-method "as:" ((number *traits-number*) (string *traits-string*))
  (format nil "~A" number))

(define-method "as:" ((string *traits-string*) (number *traits-number*))
  (values (read-from-string string)))

(define-method "as:" ((string *traits-string*) (integer *traits-integer*))
  (values (parse-integer string)))

(define-method "as:" ((character *traits-character*) (integer *traits-integer*))
  (char-code character))

(define-method "as:" ((integer *traits-integer*) (character *traits-character*))
  (code-char integer))

(define-method "as:" ((selector *traits-symbol*) (string *traits-string*))
  (symbol-name selector))

(define-method "as:" ((string *traits-string*) (selector *traits-symbol*))
  (intern string))

;; The following 4 methods are deprecated, but allow Console I/O to work with
;; just the Streams library without the ExternalResource libraries.
(define-method "next" ((stream *console-input*))
  (read-char))
(define-method "flush" ((stream *console-input*))
  (finish-output stream))
(define-method "nextPut:" ((stream *console-output*) (character *traits-character*))
  (write-char character))
(define-method "flush" ((stream *console-output*))
  (finish-output stream))

; The following methods are intended to write to Sequences of some kind, so
; that Arrays, Strings, and ByteArrays may be used equally.
(define-method "read:from:startingAt:into:" ((console *console*)
					     n handle start s)
  (read-sequence s *standard-input* :start start))

(define-method "write:to:startingAt:from:" ((console *console*)
					    n handle start s)
  (write-sequence s #-mcl *standard-output* #+mcl *terminal-io* :start start))

(define-method "flushInput" ((console *console*))
  (finish-output *standard-input*))

(define-method "flushOutput" ((console *console*))
  (finish-output #-mcl *standard-output* #+mcl *terminal-io*))

(define-method clone ((oddball *traits-oddball*))
  (if (eq oddball *primitive-oddball*)
      (clone-object *primitive-oddball*)
      oddball))

(define-method clone ((namespace *traits-namespace*))
  (if (eq namespace *primitive-namespace*)
      (clone-object *primitive-namespace*)
      namespace))

; Only here for bootstrapping the libraries.
(define-method derive ((derivable *traits-derivable*))
  (let ((object (invoke-method '|clone| nil derivable))
        (traits (clone-object *primitive-traits*)))
    (add-slot traits '|parent| (invoke-method '|traits| nil object) t t)
    (set-slot object '|traits| traits)
    object))

; Only here for bootstrapping the libraries.
(define-method "deriveWith:" ((derivable *traits-derivable*) array)
  (let ((traits (clone-object *primitive-traits*)))
    (loop for parent in (nreverse (cons derivable (coerce array 'list)))
          for index from 0
          with object
      do
        (if object
            (invoke-method '|addSlotsFrom:| nil object parent)
            (setf object (invoke-method '|clone| nil parent)))
        (add-slot traits (intern (format nil "parent~A" index))
		  (invoke-method '|traits| nil parent) t :immutable)
      finally
        (set-slot object '|traits| traits)
        (return object))))

; Only here for bootstrapping the libraries.
(define-method "allButFirst" ((array *traits-array*))
  (subseq array 1))

(define-method "addPrototype:derivedFrom:" ((obj *traits-root*) name array)
  (let* ((proto (if (eq (length array) 1)
		    (invoke-method '|derive| nil (svref array 0))
		  (invoke-method '|deriveWith:| nil
				 (svref array 0) (subseq array 1))))
	 (traits (get-slot proto '|traits|)))
    (add-slot obj name proto nil :immutable)
    (multiple-value-bind (value exists) (get-slot traits '|name|)
      (if exists
	  (set-slot traits '|name| (symbol-name name))
	(add-slot traits '|name| (symbol-name name) nil)))
    proto))

(define-method "==" ((x *traits-root*) (y *traits-root*))
  (bool-of (eq x y)))

(define-method "asMethod:on:" ((method *traits-method*) (selector *traits-symbol*) array)
  (let ((method (clone-object method)))
    (set-slot method '|selector| selector)
    (apply #'add-roles selector method (coerce array 'list))))

; Only here for bootstrapping the libraries.
(define-method "addSlotsFrom:" ((x *traits-root*) (y *traits-root*))
  (loop for slot-name in (wm-map-slot-names (wm-object-map y))
        for slot = (gethash slot-name (wm-map-slots (wm-object-map y)))
        for contents = (svref (wm-object-slots y) (wm-slot-index slot))
    do
      (add-slot x slot-name
        contents
        (wm-slot-delegates slot)
        (wm-slot-accessors slot)))
  x) 

(define-method "addRolesFrom:" ((x *traits-root*) (y *traits-root*))
  (let ((map 
          (if (wm-object-p y) 
              (wm-object-map y)
              (wm-object-map (gethash y *literal-objects*)))))
    (maphash
      #'(lambda (name slot)
          (loop for role in (wm-slot-roles slot)
                for method = (wm-role-method role)
            when (eq (get-slot method '|accessor|) *primitive-nil*)
              do (add-role x name (wm-role-positions role) (wm-role-arguments role) method)))
     (wm-map-slots map))
    x))

(define-method "addSlot:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) value nil t)
  object)

(define-method "addDelegate:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) value t t)
  object)

(define-method "addImmutableSlot:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) value nil :immutable)
  object)

(define-method "addImmutableDelegate:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) value t :immutable)
  object)

(define-method "removeSlotNamed:" ((object *traits-root*) name)
  (remove-slot object (if (symbolp name) name (intern name))))

(define-method "slotNames" ((object *traits-root*))
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*))
    (unless object (return #(|traits|))))
  (coerce (wm-map-slot-names (wm-object-map object)) 'vector))

(define-method "delegateNames" ((object *traits-root*))
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*))
    (unless object (return #(|traits|))))
  (coerce 
   (loop for slot-name in (wm-map-slot-names (wm-object-map object))
     for slot = (gethash slot-name (wm-map-slots (wm-object-map object)))
     when (wm-slot-delegates slot)
     collect slot-name)
   'vector))

(define-method clone ((object *traits-root*))
  (if (wm-object-p object)
      (clone-object object)
      (clone-object
        (or (gethash object *literal-objects*)
            (let ((representation (make-wm-object)))
              (add-slot representation '|traits| (traits-for-prim object)
			t :immutable)
              representation)))))

(define-method clone ((array *traits-byte-array*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| *traits-byte-array*
		      t :immutable)
            representation)))
    new-array))

(define-method clone ((array *traits-word-array*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| *traits-word-array*
		      t :immutable)
            representation)))
    new-array))

(define-method clone ((array *traits-array*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| *traits-array* t :immutable)
            representation)))
    new-array))

(define-method clone ((array *traits-string*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| *traits-string* t :immutable)
            representation)))
    new-array))

(define-method "newSize:" ((array *traits-byte-array*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :element-type '(unsigned-byte 8) :initial-element 0)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "newSize:" ((array *traits-word-array*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :element-type '(unsigned-byte 32) :initial-element 0)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "newSize:" ((array *traits-array*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :initial-element *primitive-nil*)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "newSize:" ((array *traits-string*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :element-type 'character :initial-element 
                #+cormanlisp #\Nul #-cormanlisp #\Null)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "atSlotIndex:" ((object *traits-root*) (index *traits-integer*))
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*)))
  (svref (wm-object-slots object) index))

(define-method "atSlotIndex:put:" ((object *traits-root*) (index *traits-integer*) value)
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*)))
  (setf (svref (wm-object-slots object) index) value))

(define-method "atSlotNamed:" ((object *traits-root*) (slot *traits-symbol*))
  (if (and (not (wm-object-p object))
           (eq slot '|traits|))
      (invoke-method '|traits| nil object)
      (get-slot object slot)))

(define-method "atSlotNamed:put:" ((object *traits-root*) (slot *traits-symbol*) value)
  (set-slot object slot value))

(define-method "at:" ((array *traits-byte-array*) (index *traits-integer*))
  (aref array index))

(define-method "at:put:" ((array *traits-byte-array*) (index *traits-integer*) value)
  (setf (aref array index) value))

(define-method size ((array *traits-byte-array*))
  (length array))

(define-method "at:" ((array *traits-word-array*) (index *traits-integer*))
  (aref array index))

(define-method "at:put:" ((array *traits-word-array*) (index *traits-integer*) value)
  (setf (aref array index) value))

(define-method size ((array *traits-word-array*))
  (length array))

(define-method "at:" ((array *traits-array*) (index *traits-integer*))
  (aref array index))

(define-method "at:put:" ((array *traits-array*) (index *traits-integer*) value)
  (setf (aref array index) value))

(define-method "includes:" ((array *traits-array*) obj)
  (bool-of (find obj array)))

(define-method size ((array *traits-array*))
  (length array))

(define-method "at:" ((string *traits-string*) (index *traits-integer*))
  (char string index))

(define-method "at:put:" ((string *traits-string*) (index *traits-integer*) value)
  (setf (char string index) value))

(define-method size ((string *traits-string*))
  (length string))

; This anticipates the loading of the Streams library.
(define-method "printOn:" ((object *traits-root*) (stream *traits-root*))
  (invoke-method '|nextPutAll:| nil stream
		 (invoke-method '|print| nil object)))

(define-method "printOn:" ((object *traits-root*) (stream *console-output*))
  (format t "~A" (invoke-method '|print| nil object)))

(define-method print ((object *traits-traits*))
  (unless (and (wm-object-p object)
               (or (eq object *traits-traits*)
                   (eq (get-slot object '|traits|) *traits-traits*)))
    (return-from print (invoke-method '|print| currentMethod object)))
  (format nil "@~A" (or (get-slot object '|name|) "_")))

(define-method print ((object *traits-symbol*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (format nil "#~A" object))

(define-method print ((object *traits-character*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (format nil "$~A"
    (case object
      (#\Newline "\\n")
      (#\Tab "\\t")
      (#\Return "\\r")
      (#\Backspace "\\b")
      (#\Space "\\s")
      (#\^[ "\\e")
      ((or #+cormanlisp #\Nul #\Null) "\\0")
      (t
        (case (char-code object)
          (7 "\\a")
          (11 "\\v")
          (12 "\\f")
          (t object))))))

(define-method print ((object *traits-root*))
  (let ((string "<"))
    (when (get-slot object '|traits|)
      (setf string
	    (format nil "~A@~A:" string
		    (or (get-slot (get-slot object '|traits|) '|name|) "_"))))
    (setf string
	  (if (< (length (wm-map-slot-names (wm-object-map object))) 20)
	      (format nil "~A ~{ ~A~}>"
		      string (wm-map-slot-names (wm-object-map object)))
	    (format nil "~A ...>" string)))
    string))

(define-method print ((object *primitive-nil*))
  "Nil")

(define-method print ((object *primitive-true*))
  "True")

(define-method print ((object *primitive-false*))
  "False")

(define-method print ((object *no-role*))
  "NoRole")

(define-method print ((file *traits-file*))
  (if (wm-object-p file)
      (invoke-method '|print| currentMethod file)
    (format nil "<File named:~A>" (namestring file))))

#+clisp
(define-method print ((socket *traits-socket-server*))
  (if (wm-object-p socket)
      (invoke-method '|print| currentMethod socket)
    (format nil "<SocketServer port:~A>" (socket:socket-server-port socket))))

#+clisp
(define-method print ((socket *traits-socket-client*))
  (if (wm-object-p socket)
      (invoke-method '|print| currentMethod socket)
    (format nil "<SocketClient port:~A>"
	    (second (multiple-value-list
		     (socket::socket-client-peer socket))))))

(define-method print ((object *traits-lisp*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (format nil "<Lambda ~A>" (function-lambda-expression object)))

(define-method print ((object *traits-number*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (if (wm-object-p object)
      "<@Number>"
      (format nil "~A" object)))

(define-method print ((array *traits-array*))
  (if (wm-object-p array)
      (invoke-method '|print| currentMethod array)
    (let ((string "{"))
      (setf string
	    (apply #'concatenate 'string string
		   (loop for index below (1- (length array))
		     collect (invoke-method '|print| nil (aref array index))
		     collect (format nil ". "))))
      (when (> (length array) 0)
	(setf string
	      (concatenate 'string string
			   (invoke-method '|print| nil (aref array (1- (length array)))))))
      (setf string (concatenate 'string string "}"))
      string)))

(define-method print ((array *traits-byte-array*))
  (if (wm-object-p array)
      (invoke-method '|print| currentMethod array)
    (let ((string "{"))
      (setf string
	    (apply #'concatenate 'string string
		   (loop for index below (1- (length array))
		     collect (format nil "~A. " (aref array index)))))
      (when (> (length array) 0)
	(setf string
	      (concatenate 'string string (format nil "~A" (aref array (1- (length array)))))))
      (setf string (concatenate 'string string "}"))
      string)))

(define-method print ((array *traits-word-array*))
  (if (wm-object-p array)
      (invoke-method '|print| currentMethod array)
    (let ((string "{"))
      (setf string
	    (apply #'concatenate 'string string
		   (loop for index below (1- (length array))
		     collect (format nil "~A. " (aref array index)))))
      (when (> (length array) 0)
	(setf string
	      (concatenate 'string string (format nil "~A" (aref array (1- (length array)))))))
      (setf string (concatenate 'string string "}"))
      string)))

(define-method print ((string *traits-string*))
  (if (wm-object-p string)
      (invoke-method '|print| currentMethod string)
    (format nil "'~A'" string)))

(define-method print ((method *traits-method*))
  (if (eq method *traits-method*)
      (invoke-method '|print| currentMethod method)
    (format nil "[~A]" (if (eq (get-slot method '|selector|) *primitive-nil*)
			   "" (get-slot method '|selector|)))))

(define-method loop ((method *traits-method*))
  (loop (apply-method method)) *primitive-nil*)

(define-method value ((method *traits-method*))
  (apply-method method))

(define-method "value:" ((method *traits-method*) x)
  (apply-method method x))

(define-method "value:value:" ((method *traits-method*) x y)
  (apply-method method x y))

(define-method "value:value:value:" ((method *traits-method*) x y z)
  (apply-method method x y z))

(define-method "values:" ((method *traits-method*) arguments &optionals optionals)
  (apply (get-slot method '|code|) method
    (if optionals
        (nconc (coerce arguments 'list) (list optionals))
        (coerce arguments 'list))))

(define-method "type" ((platform *platform*))
  (machine-type))

(define-method "version" ((platform *platform*))
  (machine-version))

(define-method "OS" ((platform *platform*))
  (software-type))

(define-method "OSversion" ((platform *platform*))
  (software-version))

(define-resource-method "newNamed:" ((file *traits-file*) (filename *traits-string*))
  (open (parse-namestring filename) :direction :io :if-exists :overwrite))

(define-resource-method "newForInputNamed:" ((file *traits-file*) (filename *traits-string*))
  (or (open (parse-namestring filename) :direction :input :if-does-not-exist nil) *primitive-nil*))

(define-resource-method "newForOutputNamed:" ((file *traits-file*) (filename *traits-string*))
  (open (parse-namestring filename) :direction :output :if-exists :overwrite))

(define-resource-method "newForNewNamed:" ((file *traits-file*) (filename *traits-string*))
  (open (parse-namestring filename)
	:direction :io :if-exists :overwrite :if-does-not-exist :create))

(define-resource-method "close" ((file *traits-file*))
  (bool-of (close file)))

(define-resource-method "fullName" ((file *traits-file*))
  (namestring (truename file)))

(define-resource-method "renameTo:" ((file *traits-file*) (newname *traits-string*))
  (multiple-value-bind (new-name old-path new-path) (rename-file file newname)
    (namestring new-path)))

(define-resource-method "delete" ((file *traits-file*))
  (bool-of (delete-file file)))

(define-resource-method "size" ((file *traits-file*))
  (file-length file))

; The following methods are intended to write to Sequences of some kind, so
; that Arrays, Strings, and ByteArrays may be used equally.
(define-resource-method "read:from:startingAt:into:" ((file *traits-file*)
						      n handle start s)
  (read-sequence s file :start start))

(define-resource-method "write:to:startingAt:from:" ((file *traits-file*)
						     n handle start s)
  (write-sequence s file :start start))

(define-resource-method "position" ((file *traits-file*))
  (file-position file))

#-cormanlisp
(define-resource-method "position:" ((file *traits-file*) pos)
  (bool-of (file-position file pos)))

(define-resource-method "size" ((file *traits-file*))
  (file-length file))

(define-resource-method "filename" ((file *traits-file*))
  (namestring file))

(define-method "debugMode" ((lobby *lobby*))
  (bool-of *slate-debuggable*))

(define-method "debugMode:" ((lobby *lobby*) (flag *traits-boolean*))
  (setf *slate-debuggable* (eq flag *primitive-true*))
  flag)

(define-method "backtrace" ((lobby *lobby*))
  (coerce
    (loop for (method selector arguments optionals) in *slate-backtrace*
          for backtrace = (clone-object *primitive-backtrace-frame*)
      do (set-slot backtrace '|method| method)
         (set-slot backtrace '|selector| selector)
         (set-slot backtrace '|arguments| arguments)
         (set-slot backtrace '|optionals| (or optionals *primitive-nil*))
      collect backtrace)
      'simple-vector))

(define-method "ensure:" ((method *traits-method*) (final *traits-method*))
  (unwind-protect (apply-method method)
    (apply-method final)))

(define-method "initHook" ((lobby *lobby*))
  *primitive-nil*)

(define-method "interpretHook" ((method *traits-method*))
  (or (apply-method method) *primitive-nil*))

(define-method "compileMode" ((lobby *lobby*))
  (bool-of *slate-compile*))

(define-method "compileMode:" ((lobby *lobby*) (flag *traits-boolean*))
  (setf *slate-compile* (eq flag *primitive-true*))
  flag)

(define-method "compileFileNamed:" ((lobby *lobby*) (filename *traits-string*))
  (let ((out-name
	 (concatenate 'string
		      (subseq filename 0 (search ".slate" filename)) ".lisp")))
    (compile-slate-to-lisp filename out-name)
    (compile-file out-name :verbose nil :print nil)
    (delete-file out-name))
  *primitive-true*)

(define-resource-method "fileIn:" ((file *traits-file*)
				   (string *traits-string*))
  (when (not (probe-file string))
    (format t "File '~A' not found. Nothing loaded." string)
    (return *primitive-nil*))
  (let ((fasl-name
	 (compile-file-pathname (subseq string 0 (search ".slate" string)))))
    (when (and (probe-file fasl-name) (not (search ".test" string)))
      (format t "~&Loading '~A'.~%" fasl-name)
      (load fasl-name :verbose nil :print nil)
      (return *primitive-nil*)))
  (invoke-method '|enterModule:| nil *lobby* string)
  (with-open-file (file string)
    (with-file-name (file-name string)
      (with-line-number (line-number 0)
	(let ((token-stream (make-token-stream :stream file)))
#-ecl
	  (loop for code = (parse-statement token-stream :eof)
	    with result = *primitive-nil*
	    do (setf result (compile-slate code))
	    if (eq (peek-token token-stream) :end-statement)
	      do (get-token token-stream)
	    else
	      if (eq (peek-token token-stream) :eof)
	      do (return result))
#+ecl
	  (funcall
	   (if *slate-compile* #'compile-slate #'interpret-slate)
	   (loop for code = (parse-statement token-stream :eof)
	     collect code into result
	     if (eq (peek-token token-stream) :end-statement)
	       do (get-token token-stream)
	     else
	       if (eq (peek-token token-stream) :eof)
	       do (return `(progn ,@(or result '(*primitive-nil*)))))))))))

; Only here for bootstrapping the libraries.
(define-method "fileIn" ((file *traits-file*))
  (invoke-method '|fileIn:| nil *primitive-file* (namestring file)))

; Only here for bootstrapping the libraries.
(define-method "fileIn" ((string *traits-string*))
  (invoke-method '|fileIn:| nil *primitive-file* string))

#+clisp
(progn
(define-resource-method "newOnPort:" ((socket *traits-socket-server*) (port *traits-integer*))
  (socket:socket-server port))
(define-resource-method "newForPeerOf:" ((socket *traits-socket-server*) (stream *traits-socket-client*))
  (socket:socket-server stream))
(define-resource-method "shutdown" ((socket *traits-socket-server*))
  (socket::socket-stream-shutdown socket))
(define-resource-method "close" ((socket *traits-socket-server*))
  (socket:socket-server-close socket))
(define-resource-method "wait" ((socket *traits-socket-server*))
  (socket:socket-wait socket))
(define-resource-method "wait:" ((socket *traits-socket-server*) (delay *traits-number*))
  (socket:socket-wait socket delay))
(define-resource-method "port" ((socket *traits-socket-server*))
  (socket:socket-server-port socket))
(define-resource-method "serverStream" ((socket *traits-socket-server*))
  (socket:socket-accept socket))
(define-resource-method "status" ((socket *traits-socket-server*))
  (socket:socket-status socket))
(define-resource-method "status" ((socket *traits-socket-client*))
  (socket:socket-status socket))
(define-resource-method "host" ((socket *traits-socket-client*))
  (socket::socket-client-peer host))
(define-resource-method "port" ((socket *traits-socket-client*))
  (second (multiple-value-list (socket::socket-client-peer socket))))
(define-resource-method "peerHost" ((socket *traits-socket-client*))
  (socket::socket-client-peer socket))
(define-resource-method "peerPort" ((socket *traits-socket-client*))
  (second (multiple-value-list (socket::socket-client-peer socket))))
(define-resource-method "peerIP" ((socket *traits-socket-client*))
  (socket::socket-client-peer socket t))
(define-resource-method "newOnPort:" ((socket *traits-socket-client*) port)
  (socket:socket-connect port))
(define-resource-method "shutdown" ((socket *traits-socket-client*))
  (socket::socket-client-shutdown socket))
(define-resource-method "close" ((socket *traits-socket-client*))
  (socket::socket-client-shutdown socket)
  (close socket))
; The following methods are intended to write to Sequences of some kind, so
; that Arrays, Strings, and ByteArrays may be used equally.
(define-resource-method "read:from:startingAt:into:" ((socket *traits-socket-client*)
					     n handle start s)
  (read-sequence s socket :start start))

(define-resource-method "write:to:startingAt:from:" ((socket *traits-socket-client*)
					    n handle start s)
  (write-sequence s socket :start start))
(define-resource-method "read:from:startingAt:into:" ((socket *traits-socket-server*)
					     n handle start s)
  (read-sequence s socket :start start))

(define-resource-method "write:to:startingAt:from:" ((socket *traits-socket-server*)
					    n handle start s)
  (write-sequence s socket :start start))
)

(define-method "findOn:" ((selector *traits-symbol*) arguments)
  (let ((arglist (coerce arguments 'list)))
    (prog1
      (dispatch-method selector nil nil arglist)
      (setf (svref arguments 0) (first arglist)))))

(define-method "findOn:after:" ((selector *traits-symbol*) arguments method)
  (let ((arglist (coerce arguments 'list)))
    (prog1
      (dispatch-method selector method nil arglist)
      (setf (svref arguments 0) (first arglist)))))

(define-method "methodNamesAt:" ((object *traits-root*) (index *traits-integer*))
  (method-names-for-position object index))

(define-method "methodsNamed:" ((object *traits-root*)
				(name *traits-symbol*))
  (find-methods object name))

(define-method "methodsNamed:at:" ((object *traits-root*)
				   (name *traits-symbol*)
				   (index *traits-integer*))
  (find-methods object name index))

(define-method "sendTo:" ((selector *traits-symbol*) arguments &optionals optionals)
  (if optionals
      (apply #'invoke-method-with-optionals selector nil optionals (coerce arguments 'list))
      (apply #'invoke-method selector nil (coerce arguments 'list))))

(define-method "sendTo:through:" ((selector *traits-symbol*) arguments dispatchers &optionals optionals)
  (let ((method (dispatch-method selector ignore t dispatchers)))
    (when optionals
      (nconc arguments optionals))
    (apply (the function (get-slot method '|code|)) method arguments)))

(define-method "sendWith:" ((selector *traits-symbol*) x)
  (invoke-method selector nil x))

(define-method "sendWith:with:" ((selector *traits-symbol*) x y)
  (invoke-method selector nil x y))

(define-method "sendWith:with:with:" ((selector *traits-symbol*) x y z)
  (invoke-method selector nil x y z))

(define-method "methodsNamed:atPosition:" ((object *traits-root*) name index)
  (let ((methods (make-array 0 :fill-pointer 0 :adjustable t)))
    (multiple-value-bind (slot exists)
	(gethash name (wm-map-slots (wm-object-map
				     (if (wm-object-p object) object
				       (or (gethash object *literal-objects*)
					   (return #()))))))
      (when exists
	(dolist (role (wm-slot-roles slot))
	  (when (logbitp index (wm-role-positions role))
	    (vector-push-extend (wm-role-method role) methods)))))
    methods))

(define-method "removeFrom:" ((method *traits-method*)
			      (objects *traits-array*))
  (let ((name (get-slot method '|selector|)))
    (loop for each across objects
      for index = 0 then (1+ index)
      for map = (wm-object-map (if (wm-object-p each) each
				  (or (gethash each *literal-objects*)
				      *no-role*)))
      for slot = (gethash name (wm-map-slots map))
      when slot
      do (setf (wm-slot-roles slot)
	       (loop for role in (wm-slot-roles slot)
		 unless (and (logbitp index (wm-role-positions role))
			     (eq method (wm-role-method role)))
		 collect role))))
  method)

(define-method "replaceWith:on:" ((method *traits-method*)
				  (new-method *traits-method*)
				  (objects *traits-array*))
  (let ((name (get-slot method '|selector|)))
    (loop for each across objects
      for index = 0 then (1+ index)
      for map = (wm-object-map (if (wm-object-p each) each
				  (or (gethash each *literal-objects*)
				      *no-role*)))
      do (loop for role in (wm-slot-roles (gethash name (wm-map-slots map)))
	   when (and (logbitp index (wm-role-positions role))
		     (eq method (wm-role-method role)))
	   do (setf (wm-role-method role) new-method))))
  method)

(define-method "identityHash" ((object *traits-root*))
  (sxhash object))

(define-method "raisedTo:" ((x *traits-number*) (y *traits-number*))
  (expt x y))

(define-method "as:" ((i *traits-integer*) (f *traits-float*))
  (coerce i 'float))

(define-method "as:" ((f *traits-float*) (i *traits-integer*))
  (values (floor f)))

(define-method "bitOr:" ((x *traits-integer*) (y *traits-integer*))
  (logior x y))

(define-method "bitXor:" ((x *traits-integer*) (y *traits-integer*))
  (logxor x y))

(define-method "bitAnd:" ((x *traits-integer*) (y *traits-integer*))
  (logand x y))

(define-method "bitShift:" ((x *traits-integer*) (y *traits-integer*))
  (if (> y 0)
      (ash x y)
      (ldb (byte (1- (integer-length x)) (- y)) x)))

(define-method "bitNot" ((x *traits-integer*))
  (lognot x))

(define-method "<<" ((x *traits-integer*) (y *traits-integer*))
  (ash x y))

(define-method ">>" ((x *traits-integer*) (y *traits-integer*))
  (ash x (- y)))

(define-method "+" ((x *traits-integer*) (y *traits-integer*))
  (+ x y))

(define-method "-" ((x *traits-integer*) (y *traits-integer*))
  (- x y))

(define-method "quo:" ((x *traits-integer*) (y *traits-integer*))
  (values (floor x y)))

(define-method "*" ((x *traits-integer*) (y *traits-integer*))
  (* x y))

(define-method "+" ((x *traits-float*) (y *traits-float*))
  (+ x y))

(define-method "-" ((x *traits-float*) (y *traits-float*))
  (- x y))

(define-method "/" ((x *traits-float*) (y *traits-float*))
  (/ x y))

(define-method "*" ((x *traits-float*) (y *traits-float*))
  (* x y))

(define-method "=" ((x *traits-integer*) (y *traits-integer*))
  (bool-of (= x y)))

(define-method ">" ((x *traits-integer*) (y *traits-integer*))
  (bool-of (> x y)))

(define-method "<" ((x *traits-integer*) (y *traits-integer*))
  (bool-of (< x y)))

(define-method ">=" ((x *traits-integer*) (y *traits-integer*))
  (bool-of (>= x y)))

(define-method "<=" ((x *traits-integer*) (y *traits-integer*))
  (bool-of (<= x y)))

(define-method "=" ((x *traits-float*) (y *traits-float*))
  (bool-of (= x y)))

(define-method ">" ((x *traits-float*) (y *traits-float*))
  (bool-of (> x y)))

(define-method "<" ((x *traits-float*) (y *traits-float*))
  (bool-of (< x y)))

(define-method ">=" ((x *traits-float*) (y *traits-float*))
  (bool-of (>= x y)))

(define-method "<=" ((x *traits-float*) (y *traits-float*))
  (bool-of (<= x y)))
 
(define-method "sin" ((x *traits-number*))
  (sin x))

(define-method "cos" ((x *traits-number*))
  (cos x))

(define-method "tan" ((x *traits-number*))
  (tan x))

(define-method "arcsin" ((x *traits-number*))
  (asin x))

(define-method "arccos" ((x *traits-number*))
  (acos x))

(define-method "arctan" ((x *traits-number*))
  (atan x))

(define-method "arctan:" ((y *traits-number*) (x *traits-number*))
  (atan y x))

#+clx (load "x11")
